
    <aside class="w-64 bg-gray-800 p-4 border-gray-700 shrink-0 rounded">
        <h2 class="text-lg font-semibold mb-4">Available Tables</h2>
        <div class="space-y-2">
            {% for table in tables %}
            <div class="table-item p-3 rounded-lg bg-gray-700 hover:bg-gray-600 cursor-pointer transition-colors"
                data-index="{{loop.index0}}" onclick="selectTable(this)">
                <div class="text-sm font-medium">{{table.name}}</div>
                <div class="text-xs text-gray-400">{{table.module}}</div>
            </div>
            {% endfor %}
        </div>
    </aside>
