<!doctype html>
<html>
  <head>
    <title>ReDoc API Reference</title>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <div id="redoc-container"></div>
    <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"></script>
    <script>
      Redoc.init({{ openapi | tojson }}, {}, document.getElementById("redoc-container"));
    </script>
  </body>
</html>
