{% extends "base.html" %}

{% block content %}
<div class="max-w-4xl mx-auto">
  <h1 class="text-2xl font-semibold mb-6">Create New Record</h1>

  <form id="createForm" class="space-y-4 bg-gray-800 p-6 rounded-lg">
    <div id="dynamicInputs" class="space-y-4">
      <!-- Dynamic inputs will be generated here -->
    </div>
    <button type="submit" class="w-full bg-blue-600 hover:bg-blue-500 text-white py-2 px-4 rounded-lg">
      Submit
    </button>
  </form>
</div>
<script> 
const isUpdate = false; 
  const data = {};
  {% include "create.js" %}
</script>
{% endblock %}
